from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import Annotated, Optional, Union

from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.events import AssetMaterialization
from dagster._core.definitions.result import MaterializeResult
from dagster_sling import DagsterSlingTranslator, SlingResource, sling_assets
from dagster_sling.resources import AssetExecutionContext
from pydantic import BaseModel
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext
from dagster_components.core.component import component_type
from dagster_components.core.component_scaffolder import ComponentScaffolder
from dagster_components.core.schema.metadata import ResolvableFieldInfo
from dagster_components.core.schema.objects import (
    AssetAttributesModel,
    AssetSpecTransformModel,
    OpSpecBaseModel,
)
from dagster_components.utils import ResolvingInfo, get_wrapped_translator_class


class SlingReplicationParams(BaseModel):
    path: str
    op: Optional[OpSpecBaseModel] = None
    asset_attributes: Annotated[
        Optional[AssetAttributesModel], ResolvableFieldInfo(additional_scope={"stream_definition"})
    ] = None


class SlingReplicationCollectionParams(BaseModel):
    sling: Optional[SlingResource] = None
    replications: Sequence[SlingReplicationParams]
    transforms: Optional[Sequence[AssetSpecTransformModel]] = None


@component_type(name="sling_replication_collection")
class SlingReplicationCollectionComponent(Component):
    def __init__(
        self,
        dirpath: Path,
        resource: SlingResource,
        sling_replications: Sequence[SlingReplicationParams],
        transforms: Sequence[AssetSpecTransformModel],
    ):
        self.dirpath = dirpath
        self.resource = resource
        self.sling_replications = sling_replications
        self.transforms = transforms

    @classmethod
    def get_scaffolder(cls) -> ComponentScaffolder:
        from dagster_components.lib.sling_replication_collection.scaffolder import (
            SlingReplicationComponentScaffolder,
        )

        return SlingReplicationComponentScaffolder()

    @classmethod
    def get_schema(cls):
        return SlingReplicationCollectionParams

    @classmethod
    def load(cls, context: ComponentLoadContext) -> Self:
        loaded_params = context.load_params(cls.get_schema())
        return cls(
            dirpath=context.path,
            resource=loaded_params.sling or SlingResource(),
            sling_replications=loaded_params.replications,
            transforms=loaded_params.transforms or [],
        )

    def build_replication_asset(
        self, context: ComponentLoadContext, replication: SlingReplicationParams
    ) -> AssetsDefinition:
        translator_cls = get_wrapped_translator_class(DagsterSlingTranslator)

        @sling_assets(
            name=replication.op.name if replication.op else Path(replication.path).stem,
            op_tags=replication.op.tags if replication.op else {},
            replication_config=self.dirpath / replication.path,
            dagster_sling_translator=translator_cls(
                base_translator=DagsterSlingTranslator(),
                resolving_info=ResolvingInfo(
                    obj_name="stream_definition",
                    asset_attributes=replication.asset_attributes or AssetAttributesModel(),
                    value_resolver=context.templated_value_resolver,
                ),
            ),
        )
        def _asset(context: AssetExecutionContext):
            yield from self.execute(context=context, sling=self.resource)

        return _asset

    def execute(
        self, context: AssetExecutionContext, sling: SlingResource
    ) -> Iterator[Union[AssetMaterialization, MaterializeResult]]:
        yield from sling.replicate(context=context)

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        defs = Definitions(
            assets=[
                self.build_replication_asset(context, replication)
                for replication in self.sling_replications
            ],
        )
        for transform in self.transforms:
            defs = transform.apply(defs, context.templated_value_resolver)
        return defs



import sys
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union, cast

import click
import typer
from pydantic import ValidationError
from pydantic_core import ErrorDetails

from dagster_components.core.component import (
    Component,
    ComponentLoadContext,
    ComponentTypeRegistry,
    TemplatedValueResolver,
    get_component_type_name,
)
from dagster_components.core.component_defs_builder import (
    component_type_from_yaml_decl,
    load_components_from_context,
    path_to_decl_node,
    resolve_decl_node_to_yaml_decls,
)
from dagster_components.core.deployment import (
    CodeLocationProjectContext,
    find_enclosing_code_location_root_path,
    is_inside_code_location_project,
)
from dagster_components.utils import CLI_BUILTIN_COMPONENT_LIB_KEY

if TYPE_CHECKING:
    from dagster._utils.source_position import SourcePosition


@click.group(name="check")
def check_cli():
    """Commands for checking components."""


def prepend_lines_with_line_numbers(
    lines_with_numbers: Sequence[tuple[Optional[int], str]],
) -> Sequence[str]:
    """Prepend each line with a line number, right-justified to the maximum line number length.

    Args:
        lines_with_numbers: A sequence of tuples, where the first element is the line number and the
            second element is the line content. Some lines may have a `None` line number, which
            will be rendered as an empty string, used for e.g. inserted error message lines.
    """
    max_line_number_length = max([len(str(n)) for n, _ in lines_with_numbers])
    return [
        f"{(str(n) if n else '').rjust(max_line_number_length)} | {line.rstrip()}"
        for n, line in lines_with_numbers
    ]


def augment_inline_error_message(location: str, msg: str):
    """Improves a subset of Pyright error messages by including location information."""
    last_location_part = location.split(".")[-1]
    if msg == "Field required":
        return f"Field `{last_location_part}` is required but not provided"
    return msg


def format_indented_error_msg(col: int, msg: str) -> str:
    """Format an error message with a caret pointing to the column where the error occurred."""
    return typer.style(" " * (col - 1) + f"^ {msg}", fg=typer.colors.YELLOW)


OFFSET_LINES_BEFORE = 2
OFFSET_LINES_AFTER = 3


def error_dict_to_formatted_error(
    component_type: Optional[type[Component]], error_details: ErrorDetails
) -> str:
    ctx = error_details.get("ctx", {})
    source_position: SourcePosition = ctx["source_position"]
    source_position_path: Sequence[SourcePosition] = ctx["source_position_path"]

    # Retrieves dotted path representation of the location of the error in the YAML file, e.g.
    # params.nested.foo.an_int
    location = cast(str, error_details["loc"])[0].split(" at ")[0]

    # Find the first source position that has a different start line than the current source position
    # This is e.g. the parent json key of the current source position
    preceding_source_position = next(
        iter(
            [
                value
                for value in reversed(list(source_position_path))
                if value.start.line < source_position.start.line
            ]
        ),
        source_position,
    )
    with open(source_position.filename) as f:
        lines = f.readlines()
        lines_with_line_numbers = list(zip(range(1, len(lines) + 1), lines))

        filtered_lines_with_line_numbers = (
            lines_with_line_numbers[
                max(
                    0, preceding_source_position.start.line - OFFSET_LINES_BEFORE
                ) : source_position.start.line
            ]
            + [
                (
                    None,
                    format_indented_error_msg(
                        source_position.start.col,
                        augment_inline_error_message(location, error_details["msg"]),
                    ),
                )
            ]
            + lines_with_line_numbers[
                source_position.start.line : source_position.end.line + OFFSET_LINES_AFTER
            ]
        )
        # Combine the filtered lines with the line numbers, and add empty lines before and after
        lines_with_line_numbers = prepend_lines_with_line_numbers(
            [(None, ""), *filtered_lines_with_line_numbers, (None, "")]
        )
        code_snippet = "\n".join(lines_with_line_numbers)

    fmt_filename = (
        f"{source_position.filename}"
        f":{typer.style(source_position.start.line, fg=typer.colors.GREEN)}"
    )
    fmt_location = typer.style(location, fg=typer.colors.BRIGHT_WHITE)
    fmt_name = typer.style(
        f"{get_component_type_name(component_type)} " if component_type else "", fg=typer.colors.RED
    )
    return f"{fmt_filename} - {fmt_name}{fmt_location} {error_details['msg']}\n{code_snippet}\n"


@check_cli.command(name="component")
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@click.pass_context
def check_component_command(ctx: click.Context, paths: Sequence[str]) -> None:
    """Check component files against their schemas, showing validation errors."""
    # Resolve paths to check against, defaulting to the current working directory
    # Any files (e.g. component yaml files) are resolved to their parent component
    # directory, so that we check the corresponding component
    path_objs = [Path(p).resolve() for p in paths] or [Path.cwd().resolve()]
    path_objs = [p.parent if p.is_file() else p for p in path_objs]

    builtin_component_lib = ctx.obj.get(CLI_BUILTIN_COMPONENT_LIB_KEY, False)
    if not is_inside_code_location_project(Path.cwd()):
        click.echo(
            click.style(
                "This command must be run inside a Dagster code location project.", fg="red"
            )
        )
        sys.exit(1)

    validation_errors: list[tuple[Union[type[Component], None], ValidationError]] = []

    context = CodeLocationProjectContext.from_code_location_path(
        find_enclosing_code_location_root_path(Path.cwd()),
        ComponentTypeRegistry.from_entry_point_discovery(
            builtin_component_lib=builtin_component_lib
        ),
    )

    for instance_path in Path(context.components_path).iterdir():
        try:
            decl_node = path_to_decl_node(path=instance_path)
        except ValidationError as e:
            validation_errors.append((None, e))
            continue

        if not decl_node:
            raise Exception(f"No component found at path {instance_path}")

        yaml_decls = resolve_decl_node_to_yaml_decls(decl_node)
        for yaml_decl in yaml_decls:
            decl_path = yaml_decl.path.resolve()
            if not any((path == decl_path or path in decl_path.parents) for path in path_objs):
                continue

            clc = ComponentLoadContext(
                resources={},
                registry=context.component_registry,
                decl_node=yaml_decl,
                templated_value_resolver=TemplatedValueResolver.default(),
            )
            try:
                load_components_from_context(clc)
            except ValidationError as e:
                component_type = component_type_from_yaml_decl(
                    context.component_registry, yaml_decl
                )
                validation_errors.append((component_type, e))

    if validation_errors:
        errs = 0
        for component_type, e in validation_errors:
            for error_dict in e.errors():
                errs += 1
                click.echo(error_dict_to_formatted_error(component_type, error_dict))

        click.echo(f"Found {errs} validation errors in {len(validation_errors)} components.")
        sys.exit(1)
    else:
        click.echo("All components validated successfully.")



import os
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Final, Optional

import tomli
from dagster._core.errors import DagsterError
from typing_extensions import Self

from dagster_components.core.component import Component, ComponentTypeRegistry
from dagster_components.utils import ensure_loadable_path, get_path_for_package

# Code location
_DEFAULT_CODE_LOCATION_COMPONENTS_LIB_SUBMODULE: Final = "lib"
_DEFAULT_CODE_LOCATION_COMPONENTS_SUBMODULE: Final = "components"


def is_inside_code_location_project(path: Path) -> bool:
    try:
        find_enclosing_code_location_root_path(path)
        return True
    except DagsterError:
        return False


def find_enclosing_code_location_root_path(path: Path) -> Path:
    """Given a path, locate the code location root directory that contains it. It
    determines this by finding a pyproject.toml with a [tool.dagster] section.

    Searches parent directory recursively until it finds it. It if navigates
    to the root directory, it throws an error.
    """
    current_path = path.absolute()
    while not _is_code_location_root(current_path):
        current_path = current_path.parent
        if str(current_path) == "/":
            raise DagsterError("Cannot find code location root")
    return current_path


def _is_code_location_root(path: Path) -> bool:
    if (path / "pyproject.toml").exists():
        with open(path / "pyproject.toml") as f:
            toml = tomli.loads(f.read())
            return bool(toml.get("tool", {}).get("dagster"))
    return False


@dataclass
class CodeLocationConfig:
    component_package: Optional[str] = None


def _load_code_location_config(path: Path) -> CodeLocationConfig:
    with open(path / "pyproject.toml") as f:
        toml = tomli.loads(f.read())
        raw_dg_config = toml.get("tool", {}).get("dg", {})
        raw_config = {
            k: v for k, v in raw_dg_config.items() if k in CodeLocationConfig.__annotations__
        }
        return CodeLocationConfig(**raw_config)


class CodeLocationProjectContext:
    @classmethod
    def from_code_location_path(
        cls,
        path: Path,
        component_registry: "ComponentTypeRegistry",
        components_path: Optional[Path] = None,
        components_package_name: Optional[str] = None,
    ) -> Self:
        if not _is_code_location_root(path):
            raise DagsterError(
                f"Path {path} is not a code location root. Must have a pyproject.toml with a [tool.dagster] section."
            )

        name = os.path.basename(path)
        root_package = name.replace("-", "_")
        config = _load_code_location_config(path)
        components_package_name = (
            config.component_package
            or f"{root_package}.{_DEFAULT_CODE_LOCATION_COMPONENTS_SUBMODULE}"
        )
        if not components_path:
            with ensure_loadable_path(path):
                components_path = (
                    Path(get_path_for_package(config.component_package))
                    if config.component_package
                    else path / root_package / _DEFAULT_CODE_LOCATION_COMPONENTS_SUBMODULE
                )

        return cls(
            root_path=str(path),
            name=name,
            root_package=root_package,
            component_registry=component_registry,
            components_path=components_path,
            components_package_name=components_package_name,
        )

    def __init__(
        self,
        root_path: str,
        name: str,
        root_package: str,
        component_registry: "ComponentTypeRegistry",
        components_path: Path,
        components_package_name: str,
    ):
        self._root_path = root_path
        self._name = name
        self._root_package = root_package
        self._component_registry = component_registry
        self._components_path = components_path
        self._components_package_name = components_package_name

    @property
    def component_registry(self) -> "ComponentTypeRegistry":
        return self._component_registry

    def has_component_type(self, name: str) -> bool:
        return self._component_registry.has(name)

    def get_component_type(self, name: str) -> type[Component]:
        if not self.has_component_type(name):
            raise DagsterError(f"No component type named {name}")
        return self._component_registry.get(name)

    def list_component_types(self) -> Iterable[tuple[str, type[Component]]]:
        for key in sorted(self._component_registry.keys()):
            yield key, self._component_registry.get(key)

    def get_component_instance_path(self, name: str) -> str:
        if name not in self.component_instances:
            raise DagsterError(f"No component instance named {name}")
        return os.path.join(self.components_path, name)

    @property
    def components_package_name(self) -> str:
        return self._components_package_name

    @property
    def components_path(self) -> str:
        return str(self._components_path)

    @property
    def component_instances(self) -> Iterable[str]:
        return sorted(os.listdir(self.components_path))

    def has_component_instance(self, name: str) -> bool:
        return os.path.exists(os.path.join(self.components_path, name))



from collections.abc import Iterator, Sequence
from typing import Annotated, Optional

from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster_dbt import (
    DagsterDbtTranslator,
    DbtCliResource,
    DbtManifestAssetSelection,
    DbtProject,
    dbt_assets,
)
from pydantic import BaseModel
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext
from dagster_components.core.component import component_type
from dagster_components.core.schema.metadata import ResolvableFieldInfo
from dagster_components.core.schema.objects import (
    AssetAttributesModel,
    AssetSpecTransformModel,
    OpSpecBaseModel,
    TemplatedValueResolver,
)
from dagster_components.lib.dbt_project.scaffolder import DbtProjectComponentScaffolder
from dagster_components.utils import ResolvingInfo, get_wrapped_translator_class


class DbtProjectParams(BaseModel):
    dbt: DbtCliResource
    op: Optional[OpSpecBaseModel] = None
    asset_attributes: Annotated[
        Optional[AssetAttributesModel], ResolvableFieldInfo(additional_scope={"node"})
    ] = None
    transforms: Optional[Sequence[AssetSpecTransformModel]] = None


@component_type(name="dbt_project")
class DbtProjectComponent(Component):
    def __init__(
        self,
        dbt_resource: DbtCliResource,
        op_spec: Optional[OpSpecBaseModel],
        asset_attributes: Optional[AssetAttributesModel],
        transforms: Sequence[AssetSpecTransformModel],
        value_resolver: TemplatedValueResolver,
    ):
        self.dbt_resource = dbt_resource
        self.project = DbtProject(self.dbt_resource.project_dir)
        self.op_spec = op_spec
        self.asset_attributes = asset_attributes
        self.transforms = transforms
        self.value_resolver = value_resolver
        self.translator = self._get_wrapped_translator()

    @classmethod
    def get_scaffolder(cls) -> "DbtProjectComponentScaffolder":
        return DbtProjectComponentScaffolder()

    @classmethod
    def get_schema(cls):
        return DbtProjectParams

    @classmethod
    def load(cls, context: ComponentLoadContext) -> Self:
        loaded_params = context.load_params(cls.get_schema())

        return cls(
            dbt_resource=loaded_params.dbt,
            op_spec=loaded_params.op,
            asset_attributes=loaded_params.asset_attributes,
            transforms=loaded_params.transforms or [],
            value_resolver=context.templated_value_resolver,
        )

    def get_translator(self) -> DagsterDbtTranslator:
        return DagsterDbtTranslator()

    def _get_wrapped_translator(self) -> DagsterDbtTranslator:
        translator_cls = get_wrapped_translator_class(DagsterDbtTranslator)
        return translator_cls(
            base_translator=self.get_translator(),
            resolving_info=ResolvingInfo(
                obj_name="node",
                asset_attributes=self.asset_attributes or AssetAttributesModel(),
                value_resolver=self.value_resolver,
            ),
        )

    def get_asset_selection(
        self, select: str, exclude: Optional[str] = None
    ) -> DbtManifestAssetSelection:
        return DbtManifestAssetSelection.build(
            manifest=self.project.manifest_path,
            dagster_dbt_translator=self.translator,
            select=select,
            exclude=exclude,
        )

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        self.project.prepare_if_dev()

        @dbt_assets(
            manifest=self.project.manifest_path,
            project=self.project,
            name=self.op_spec.name if self.op_spec else self.project.name,
            op_tags=self.op_spec.tags if self.op_spec else None,
            dagster_dbt_translator=self.translator,
        )
        def _fn(context: AssetExecutionContext):
            yield from self.execute(context=context, dbt=self.dbt_resource)

        defs = Definitions(assets=[_fn])
        for transform in self.transforms:
            defs = transform.apply(defs, context.templated_value_resolver)
        return defs

    def execute(self, context: AssetExecutionContext, dbt: DbtCliResource) -> Iterator:
        yield from dbt.cli(["build"], context=context).stream()



import sys
from pathlib import Path
from typing import Optional

import click
from pydantic import TypeAdapter

from dagster_components import ComponentTypeRegistry
from dagster_components.core.deployment import (
    CodeLocationProjectContext,
    find_enclosing_code_location_root_path,
    is_inside_code_location_project,
)
from dagster_components.scaffold import (
    ComponentScaffolderUnavailableReason,
    scaffold_component_instance,
)
from dagster_components.utils import CLI_BUILTIN_COMPONENT_LIB_KEY


@click.group(name="scaffold")
def scaffold_cli() -> None:
    """Commands for scaffolding Dagster components and related entities."""


@scaffold_cli.command(name="component")
@click.argument("component_type", type=str)
@click.argument("component_name", type=str)
@click.option("--json-params", type=str, default=None)
@click.pass_context
def scaffold_component_command(
    ctx: click.Context,
    component_type: str,
    component_name: str,
    json_params: Optional[str],
) -> None:
    builtin_component_lib = ctx.obj.get(CLI_BUILTIN_COMPONENT_LIB_KEY, False)
    if not is_inside_code_location_project(Path.cwd()):
        click.echo(
            click.style(
                "This command must be run inside a Dagster code location project.", fg="red"
            )
        )
        sys.exit(1)

    context = CodeLocationProjectContext.from_code_location_path(
        find_enclosing_code_location_root_path(Path.cwd()),
        ComponentTypeRegistry.from_entry_point_discovery(
            builtin_component_lib=builtin_component_lib
        ),
    )
    if not context.has_component_type(component_type):
        click.echo(
            click.style(f"No component type `{component_type}` could be resolved.", fg="red")
        )
        sys.exit(1)

    component_type_cls = context.get_component_type(component_type)
    if json_params:
        scaffolder = component_type_cls.get_scaffolder()
        if isinstance(scaffolder, ComponentScaffolderUnavailableReason):
            raise Exception(
                f"Component type {component_type} does not have a scaffolder. Reason: {scaffolder.message}."
            )
        scaffold_params = TypeAdapter(scaffolder.get_params_schema_type()).validate_json(
            json_params
        )
    else:
        scaffold_params = {}

    scaffold_component_instance(
        context.components_path,
        component_name,
        component_type_cls,
        component_type,
        scaffold_params,
    )



from collections.abc import Mapping, Sequence
from typing import Annotated, Any, Literal, Optional, Union

import dagster._check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.asset_spec import AssetSpec, map_asset_specs
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.definitions_class import Definitions
from dagster._record import replace

from dagster_components.core.schema.base import ComponentSchemaBaseModel
from dagster_components.core.schema.metadata import ResolvableFieldInfo
from dagster_components.core.schema.resolver import TemplatedValueResolver


def _post_process_key(resolved: Optional[str]) -> Optional[AssetKey]:
    return AssetKey.from_user_string(resolved) if resolved else None


class OpSpecBaseModel(ComponentSchemaBaseModel):
    name: Optional[str] = None
    tags: Optional[dict[str, str]] = None


class AssetAttributesModel(ComponentSchemaBaseModel):
    key: Annotated[
        Optional[str],
        ResolvableFieldInfo(output_type=AssetKey, post_process_fn=_post_process_key),
    ] = None
    deps: Sequence[str] = []
    description: Optional[str] = None
    metadata: Annotated[
        Union[str, Mapping[str, Any]], ResolvableFieldInfo(output_type=Mapping[str, Any])
    ] = {}
    group_name: Optional[str] = None
    skippable: bool = False
    code_version: Optional[str] = None
    owners: Sequence[str] = []
    tags: Annotated[
        Union[str, Mapping[str, str]], ResolvableFieldInfo(output_type=Mapping[str, str])
    ] = {}
    kinds: Optional[Sequence[str]] = None
    automation_condition: Annotated[
        Optional[str], ResolvableFieldInfo(output_type=Optional[AutomationCondition])
    ] = None


class AssetSpecTransformModel(ComponentSchemaBaseModel):
    target: str = "*"
    operation: Literal["merge", "replace"] = "merge"
    attributes: AssetAttributesModel

    class Config:
        arbitrary_types_allowed = True

    def apply_to_spec(
        self,
        spec: AssetSpec,
        value_resolver: TemplatedValueResolver,
    ) -> AssetSpec:
        # add the original spec to the context and resolve values
        attributes = self.attributes.resolve_properties(value_resolver.with_scope(asset=spec))

        if self.operation == "merge":
            mergeable_attributes = {"metadata", "tags"}
            merge_attributes = {k: v for k, v in attributes.items() if k in mergeable_attributes}
            replace_attributes = {
                k: v for k, v in attributes.items() if k not in mergeable_attributes
            }
            return spec.merge_attributes(**merge_attributes).replace_attributes(
                **replace_attributes
            )
        elif self.operation == "replace":
            return spec.replace_attributes(**attributes)
        else:
            check.failed(f"Unsupported operation: {self.operation}")

    def apply(self, defs: Definitions, value_resolver: TemplatedValueResolver) -> Definitions:
        target_selection = AssetSelection.from_string(self.target, include_sources=True)
        target_keys = target_selection.resolve(defs.get_asset_graph())

        mappable = [d for d in defs.assets or [] if isinstance(d, (AssetsDefinition, AssetSpec))]
        mapped_assets = map_asset_specs(
            lambda spec: self.apply_to_spec(spec, value_resolver)
            if spec.key in target_keys
            else spec,
            mappable,
        )

        assets = [
            *mapped_assets,
            *[d for d in defs.assets or [] if not isinstance(d, (AssetsDefinition, AssetSpec))],
        ]
        return replace(defs, assets=assets)



from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Optional, Union

from dagster._record import record
from dagster._utils.pydantic_yaml import _parse_and_populate_model_with_annotated_errors
from dagster._utils.source_position import SourcePositionTree
from dagster._utils.yaml_utils import parse_yaml_with_source_positions
from pydantic import BaseModel

from dagster_components.core.component import ComponentDeclNode


class ComponentFileModel(BaseModel):
    type: str
    params: Optional[Mapping[str, Any]] = None


@record
class YamlComponentDecl(ComponentDeclNode):
    path: Path
    component_file_model: ComponentFileModel
    source_position_tree: Optional[SourcePositionTree] = None

    @staticmethod
    def from_path(component_file_path: Path) -> "YamlComponentDecl":
        parsed = parse_yaml_with_source_positions(
            component_file_path.read_text(), str(component_file_path)
        )
        obj = _parse_and_populate_model_with_annotated_errors(
            cls=ComponentFileModel, obj_parse_root=parsed, obj_key_path_prefix=[]
        )

        return YamlComponentDecl(
            path=component_file_path.parent,
            component_file_model=obj,
            source_position_tree=parsed.source_position_tree,
        )


@record
class ComponentFolder(ComponentDeclNode):
    path: Path
    sub_decls: Sequence[Union[YamlComponentDecl, "ComponentFolder"]]


def path_to_decl_node(path: Path) -> Optional[ComponentDeclNode]:
    # right now, we only support two types of components, both of which are folders
    # if the folder contains a component.yaml file, it's a component instance
    # otherwise, it's a folder containing sub-components

    if not path.is_dir():
        return None

    component_path = path / "component.yaml"

    if component_path.exists():
        return YamlComponentDecl.from_path(component_path)

    subs = []
    for subpath in path.iterdir():
        component = path_to_decl_node(subpath)
        if component:
            subs.append(component)

    return ComponentFolder(path=path, sub_decls=subs)



import json
from pathlib import Path
from typing import Any

import click

from dagster_components.core.component import (
    Component,
    ComponentTypeMetadata,
    ComponentTypeRegistry,
)
from dagster_components.core.deployment import (
    CodeLocationProjectContext,
    find_enclosing_code_location_root_path,
    is_inside_code_location_project,
)
from dagster_components.utils import CLI_BUILTIN_COMPONENT_LIB_KEY


@click.group(name="list")
def list_cli():
    """Commands for listing Dagster components and related entities."""


@list_cli.command(name="component-types")
@click.pass_context
def list_component_types_command(ctx: click.Context) -> None:
    """List registered Dagster components."""
    builtin_component_lib = ctx.obj.get(CLI_BUILTIN_COMPONENT_LIB_KEY, False)
    output: dict[str, Any] = {}
    if not is_inside_code_location_project(Path.cwd()):
        registry = ComponentTypeRegistry.from_entry_point_discovery(
            builtin_component_lib=builtin_component_lib
        )
        for key in sorted(registry.keys()):
            _add_component_type_to_output(output, key, registry.get(key))
    else:
        context = CodeLocationProjectContext.from_code_location_path(
            find_enclosing_code_location_root_path(Path.cwd()),
            ComponentTypeRegistry.from_entry_point_discovery(
                builtin_component_lib=builtin_component_lib
            ),
        )
        for key, component_type in context.list_component_types():
            _add_component_type_to_output(output, key, component_type)

    click.echo(json.dumps(output))


def _add_component_type_to_output(
    output: dict[str, Any], key: str, component_type: type[Component]
) -> None:
    package, name = key.rsplit(".", 1)
    output[key] = ComponentTypeMetadata(
        name=name,
        package=package,
        **component_type.get_metadata(),
    )



import importlib.util
import sys
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.errors import DagsterError

from dagster_components.core.schema.objects import AssetAttributesModel
from dagster_components.core.schema.resolver import TemplatedValueResolver

CLI_BUILTIN_COMPONENT_LIB_KEY = "builtin_component_lib"


def ensure_dagster_components_tests_import() -> None:
    from dagster_components import __file__ as dagster_components_init_py

    dagster_components_package_root = (Path(dagster_components_init_py) / ".." / "..").resolve()
    assert (
        dagster_components_package_root / "dagster_components_tests"
    ).exists(), "Could not find dagster_components_tests where expected"
    sys.path.append(dagster_components_package_root.as_posix())


# Temporarily places a path at the front of sys.path, ensuring that any modules in that path are
# importable.
@contextmanager
def ensure_loadable_path(path: Path) -> Iterator[None]:
    orig_path = sys.path.copy()
    sys.path.insert(0, str(path))
    try:
        yield
    finally:
        sys.path = orig_path


def get_path_for_package(package_name: str) -> str:
    spec = importlib.util.find_spec(package_name)
    if not spec:
        raise DagsterError(f"Cannot find package: {package_name}")
    # file_path = spec.origin
    submodule_search_locations = spec.submodule_search_locations
    if not submodule_search_locations:
        raise DagsterError(f"Package does not have any locations for submodules: {package_name}")
    return submodule_search_locations[0]


@dataclass
class ResolvingInfo:
    obj_name: str
    asset_attributes: AssetAttributesModel
    value_resolver: TemplatedValueResolver

    def get_resolved_attribute(self, attribute: str, obj: Any, default_method) -> Any:
        renderer = self.value_resolver.with_scope(**{self.obj_name: obj})
        rendered_attributes = self.asset_attributes.resolve_properties(renderer)
        return (
            rendered_attributes[attribute]
            if attribute in rendered_attributes
            else default_method(obj)
        )

    def get_asset_spec(self, base_spec: AssetSpec, context: Mapping[str, Any]) -> AssetSpec:
        """Returns an AssetSpec that combines the base spec with attributes resolved using the provided context.

        Usage:

        ```python
        class WrappedDagsterXTranslator(DagsterXTranslator):
            def __init__(self, *, base_translator, resolving_info: ResolvingInfo):
                self.base_translator = base_translator
                self.resolving_info = resolving_info

            def get_asset_spec(self, base_spec: AssetSpec, x_params: Any) -> AssetSpec:
                return self.resolving_info.get_asset_spec(
                    base_spec, {"x_params": x_params}
                )

        ```
        """
        resolver = self.value_resolver.with_scope(**context)
        resolved_attributes = self.asset_attributes.resolve_properties(resolver)
        return base_spec.replace_attributes(**resolved_attributes)


def get_wrapped_translator_class(translator_type: type):
    """Temporary hack to allow wrapping of many methods of a given translator class. Will be removed
    once all translators implement `get_asset_spec`.
    """

    class WrappedTranslator(translator_type):
        def __init__(self, *, base_translator, resolving_info: ResolvingInfo):
            self.base_translator = base_translator
            self.resolving_info = resolving_info

        def get_asset_key(self, obj: Any) -> AssetKey:
            return self.resolving_info.get_resolved_attribute(
                "key", obj, self.base_translator.get_asset_key
            )

        def get_group_name(self, obj: Any) -> Optional[str]:
            return self.resolving_info.get_resolved_attribute(
                "group_name", obj, self.base_translator.get_group_name
            )

        def get_tags(self, obj: Any) -> Mapping[str, str]:
            return self.resolving_info.get_resolved_attribute(
                "tags", obj, self.base_translator.get_tags
            )

        def get_automation_condition(self, obj: Any) -> Optional[AutomationCondition]:
            return self.resolving_info.get_resolved_attribute(
                "automation_condition", obj, self.base_translator.get_automation_condition
            )

    return WrappedTranslator



import functools
import os
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Optional, TypeVar, Union

from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._record import record
from jinja2.nativetypes import NativeTemplate
from pydantic import BaseModel

from dagster_components.core.schema.metadata import allow_resolve

T = TypeVar("T")


def env_scope(key: str) -> Optional[str]:
    return os.environ.get(key)


def automation_condition_scope() -> Mapping[str, Any]:
    return {
        "eager": AutomationCondition.eager,
        "on_cron": AutomationCondition.on_cron,
    }


ShouldResolveFn = Callable[[Sequence[Union[str, int]]], bool]


@record
class TemplatedValueResolver:
    scope: Mapping[str, Any]

    @staticmethod
    def default() -> "TemplatedValueResolver":
        return TemplatedValueResolver(
            scope={"env": env_scope, "automation_condition": automation_condition_scope()}
        )

    def with_scope(self, **additional_scope) -> "TemplatedValueResolver":
        return TemplatedValueResolver(scope={**self.scope, **additional_scope})

    def _resolve_value(self, val: Any) -> Any:
        """Resolves a single value, if it is a templated string."""
        return NativeTemplate(val).render(**self.scope) if isinstance(val, str) else val

    def _resolve_obj(
        self,
        obj: Any,
        valpath: Optional[Sequence[Union[str, int]]],
        should_render: Callable[[Sequence[Union[str, int]]], bool],
    ) -> Any:
        """Recursively resolves templated values in a nested object, based on the provided should_render function."""
        if valpath is not None and not should_render(valpath):
            return obj
        elif isinstance(obj, dict):
            # render all values in the dict
            return {
                k: self._resolve_obj(
                    v, [*valpath, k] if valpath is not None else None, should_render
                )
                for k, v in obj.items()
            }
        elif isinstance(obj, list):
            # render all values in the list
            return [
                self._resolve_obj(v, [*valpath, i] if valpath is not None else None, should_render)
                for i, v in enumerate(obj)
            ]
        else:
            return self._resolve_value(obj)

    def resolve_obj(self, val: Any) -> Any:
        """Recursively resolves templated values in a nested object."""
        return self._resolve_obj(val, None, lambda _: True)

    def resolve_params(self, val: T, target_type: type) -> T:
        """Given a raw params value, preprocesses it by rendering any templated values that are not marked as deferred in the target_type's json schema."""
        json_schema = (
            target_type.model_json_schema() if issubclass(target_type, BaseModel) else None
        )
        if json_schema is None:
            should_render = lambda _: True
        else:
            should_render = functools.partial(
                allow_resolve, json_schema=json_schema, subschema=json_schema
            )
        return self._resolve_obj(val, [], should_render=should_render)
